import re

from .chars import (
    ANYSP,
    DLQUO,
    DPRIME,
    HYPHEN,
    HYPHEN_MINUS,
    LAQUO,
    LDQUO,
    LSQUO,
    MDASH,
    MDASH_PAIR,
    MINUS,
    NBHYPHEN,
    NBSP,
    NDASH,
    NNBSP,
    RAQUO,
    RDQUO,
    RSQUO,
    SPRIME,
    TIMES,
    WHSP,
)
from .constants import (
    CURRENCY_SYMBOLS,
    CURRENCY_SYMBOLS_USED_IN_FORMULAS,
    POLISH_ZLOTY,
    RU_CUSTOM_REPLACEMENTS,
    RU_LONG_PREPOSITIONS,
    RU_PARTICLES,
    RU_VERBOSE_PREPOSITIONS,
)
from .utils import map_choices, without

__all__ = ('EnQuotes', 'RuQuotes', 'EnRuExpressions', 'RuExpressions')


class EnQuotes:
    r"""
    Provides English quotes configutation for :class:`typus.processors.Quotes`
    processor.

    >>> en_typus('He said "\'Winnie-the-Pooh\' is my favorite book!".')
    'He said “‘Winnie-the-Pooh’ is my favorite book!”.'
    """

    # Left odd, right odd, left even, right even
    loq, roq, leq, req = LDQUO, RDQUO, LSQUO, RSQUO


class RuQuotes:
    r"""
    Provides Russian quotes configutation for :class:`typus.processors.Quotes`
    processor.

    >>> ru_typus('Он сказал: "\'Винни-Пух\' -- моя любимая книга!".')
    'Он сказал: «„Винни-Пух“ — моя любимая книга!».'
    """

    # Left odd, right odd, left even, right even
    loq, roq, leq, req = LAQUO, RAQUO, DLQUO, LDQUO


class EnRuExpressions:
    """
    This class holds most of Typus functionality for English and Russian
    languages. It works with :class:`typus.processors.Expressions`.
    """

    expressions = (
        'spaces',
        'linebreaks',
        'apostrophe',
        'complex_symbols',
        'mdash',
        'primes',
        # 'phones',
        'digit_spaces',
        'pairs',
        'vulgar_fractions',
        # 'math',
        # 'ruble',
        'rep_positional_spaces',
        'del_positional_spaces',
        'rep_hyphen_between_short_words',
    )

    # Any unicode word
    words = r'[^\W\d_]'

    complex_symbols = {
        '...': '…',
        '<-': '←',
        '->': '→',
        '+-': '±',
        '+' + MINUS: '±',
        '<=': '≤',
        '>=': '≥',
        '/=': '≠',
        # Removed, as breaks tokens
        # '==': '≡',
        '(r)': '®',
        '(c)': '©',
        '(p)': '℗',
        '(tm)': '™',
        '(sm)': '℠',
        # cyrillic
        '(с)': '©',
        '(р)': '℗',
        '(тм)': '™',
    }

    vulgar_fractions = {
        '1/2': '½',
        '1/3': '⅓',
        '1/5': '⅕',
        '1/6': '⅙',
        '1/8': '⅛',
        '2/3': '⅔',
        '2/5': '⅖',
        '3/4': '¾',
        '3/5': '⅗',
        '3/8': '⅜',
        '4/5': '⅘',
        '5/6': '⅚',
        '5/8': '⅝',
        '7/8': '⅞',
    }

    math = {
        '-': MINUS,
        # '*xх': TIMES,
    }

    # Not need to put >=, +-, etc, after expr_complex_symbols
    math_operators = fr'[\-{MINUS}\*xх{TIMES}\+\=±≤≥≠÷\/]'

    rep_positional_spaces = {
        # No need to put vulgar fractions in here because of expr_digit_spaces
        # which joins digits and words afterward
        'after': f'←%±{MINUS}{TIMES}©§¶№',
        'both': '&≡≤≥≠',
        'before': f'→{MDASH}',
    }

    del_positional_spaces = {'before': '®℗™℠:,.?!…'}

    # Replace this if you don't need nbsp before ruble
    ruble = NBSP + '₽'

    def expr_spaces(self):
        """
        Trims spaces at the beginning and end of the line and remove extra
        spaces within.

        >>> en_typus('   foo bar  ')
        'foo bar'

        .. caution::
            Doesn't work correctly with nbsp (replaces with whitespace).
        """

        expr = ((fr'{ANYSP}{{2,}}', WHSP), (r'(?:^{0}+|{0}+$)'.format(ANYSP), ''))
        return expr

    def expr_linebreaks(self):
        r"""
        Converts line breaks to unix-style and removes extra breaks
        if found more than two in a row.

        >>> en_typus('foo\r\nbar\n\n\nbaz')
        'foo\nbar\n\nbaz'
        """

        expr = ((r'\r\n', '\n'), (r'\n{2,}', '\n' * 2))
        return expr

    def expr_apostrophe(self):
        """
        Replaces single quote with apostrophe.

        >>> en_typus("She'd, I'm, it's, don't, you're, he'll, 90's")
        'She’d, I’m, it’s, don’t, you’re, he’ll, 90’s'

        .. note::
            By the way it works with any omitted word. But then again, why not?
        """

        expr = ((r'(?<={0}|[0-9])\'(?={0})'.format(self.words), RSQUO),)
        return expr

    def expr_complex_symbols(self):
        """
        Replaces complex symbols with Unicode characters. Doesn't care
        about case-sensitivity and handles Cyrillic-Latin twins
        like ``c`` and ``с``.

        >>> en_typus('(c)(с)(C)(r)(R)...')
        '©©©®®…'

        .. csv-table:: Character map
            :header: …, ←, →, ±, ≤, ≥, ≠, ≡, ®, ©, ℗, ™, ℠

            ..., <-, ->, +- or +−, <=, >=, /=, ==, (r), (c), (p), (tm), (sm)
        """

        expr = (map_choices(self.complex_symbols),)
        return expr

    def expr_mdash(self):
        """
        Replaces dash with mdash.

        >>> en_typus('foo -- bar')  # adds non-breakable space after `foo`
        'foo\u00A0— bar'
        """

        expr = (
            # Double dash guarantees to be replaced with mdash
            (r'{0}--{0}'.format(WHSP), MDASH_PAIR),
            # Dash can be between anything except digits
            # because in that case it's not obvious
            (r'{0}+[\-|{1}]{0}+(?!\d\b)'.format(ANYSP, NDASH), MDASH_PAIR),
            # Same but backwards
            # It joins non-digit with digit or word
            (r'(\b\D+){0}+[\-|{1}]{0}+'.format(ANYSP, NDASH), fr'\1{MDASH_PAIR}'),
            # Line beginning adds nbsp after dash
            (fr'^\-{{1,2}}{ANYSP}+', fr'{MDASH}{NBSP}'),
            # Also mdash can be at the end of the line in poems
            (r'{0}+\-{{1,2}}{0}*(?=$|<br/?>)'.format(ANYSP), fr'{NBSP}{MDASH}'),
        )
        return expr

    def expr_primes(self):
        r"""
        Replaces quotes with prime after digits.

        >>> en_typus('3\' 5" long')
        '3′ 5″ long'

        .. caution::
            Won't break "4", but fails with " 4".
        """

        expr = (
            (fr'(^|{ANYSP})(\d+)\'', r'\1\2' + SPRIME),
            (fr'(^|{ANYSP})(\d+)"', r'\1\2' + DPRIME),
        )
        return expr

    def expr_phones(self):
        """
        Replaces dash with ndash in phone numbers which should be a trio of
        2-4 length digits.

        >>> en_typus('111-00-00'), en_typus('00-111-00'), en_typus('00-00-111')
        ('111–00–00', '00–111–00', '00–00–111')
        """

        expr = (
            (
                r'([0-9]{2,4})\-([0-9]{2,4})\-([0-9]{2,4})',
                r'\1{0}\2{0}\3'.format(NDASH),
            ),
        )
        return expr

    def expr_digit_spaces(self):
        """
        Replaces whitespace with non-breakable space after 4 (and less)
        length digits if word or digit without comma or math operators
        found afterwards:
        3 apples
        40 000 bucks
        400 + 3
        Skips:
        4000 bucks
        40 000,00 bucks
        """

        expr = (
            (
                r'\b(\d{{1,3}}){0}(?=[0-9]+\b|{1}|{2})'.format(
                    WHSP, self.words, self.math_operators
                ),
                r'\1' + NBSP,
            ),
        )
        return expr

    def expr_pairs(self):
        """
        Replaces whitespace with non-breakable space after 1-2 length words.
        """

        expr = (
            # Unions, units and all that small staff
            (fr'\b({self.words}{{1,2}}){WHSP}+', r'\1' + NBSP),
            # Fixes previous with leading dash, ellipsis or apostrophe
            (fr'([-…’]{self.words}{{1,2}}){NBSP}', r'\1' + WHSP),
        )
        return expr

    def expr_units(self):
        """
        Puts non-breakable space between digits and units.

        >>> en_typus('1mm', debug=True), en_typus('1mm')
        ('1_mm', '1 mm')
        """

        expr = (
            (
                fr'\b(\d+){WHSP}*(?!(?:nd|rd|th|d|g|px)\b)({self.words}{{1,3}})\b',
                fr'\1{NBSP}\2',
            ),
        )
        return expr

    def expr_ranges(self):
        """
        Replaces dash with mdash in ranges.
        Supports float and negative values.
        Tries to not mess with minus: skips if any math operator or word
        was found after dash: 3-2=1, 24-pin.
        **NOTE**: _range_ should not have spaces between dash: `2-3` and
        left side should be less than right side.
        """

        def ufloat(string):
            return float(string.replace(',', '.'))

        def replace(match):
            left, dash, right = match.groups()
            if ufloat(left) < ufloat(right):
                dash = MDASH
            return f'{left}{dash}{right}'

        expr = (
            (
                r'(-?(?:[0-9]+[\.,][0-9]+|[0-9]+))(-)'
                r'([0-9]+[\.,][0-9]+|[0-9]+)'
                r'(?!{}+{}|{})'.format(ANYSP, self.math_operators, self.words),
                replace,
            ),
        )
        return expr

    def expr_vulgar_fractions(self):
        """
        Replaces vulgar fractions with appropriate unicode characters.

        >>> en_typus('1/2')
        '½'
        """

        expr = (
            # \b to excludes digits which are not on map, like `11/22`
            map_choices(self.vulgar_fractions, r'\b({0})\b'),
        )
        return expr

    def expr_math(self):
        """
        Puts minus and multiplication symbols between pair and before
        single digits.

        >>> en_typus('3 - 3 = 0')
        '3 − 3 = 0'
        >>> en_typus('-3 degrees')
        '−3 degrees'
        >>> en_typus('3 x 3 = 9')
        '3 × 3 = 9'
        >>> en_typus('x3 better!')
        '×3 better!'

        .. important::

            Should run after `mdash` and `phones` expressions.
        """

        expr = (
            (r'(^|\d{0}*|{0}*)[{1}]({0}*\d)'.format(ANYSP, re.escape(x)), fr'\1{y}\2')
            for x, y in self.math.items()
        )
        return expr

    def expr_abbrs(self):
        """
        Adds narrow non-breakable space and replaces whitespaces between
        shorten words.
        """

        expr = (
            (r'\b({1}\.){0}*({1}\.)'.format(ANYSP, self.words), fr'\1{NNBSP}\2'),
            (r'\b({1}\.){0}*(?={1})'.format(WHSP, self.words), fr'\1{NBSP}'),
        )
        return expr

    def expr_ruble(self):
        """
        Replaces `руб` and `р` (with or without dot) after digits
        with ruble symbol.

        >>> en_typus('1000 р.')
        '1000 ₽'

        .. caution::

            Drops the dot at the end of sentence if match found in there.
        """

        expr = ((fr'(\d){ANYSP}*(?:руб|р)\b\.?', fr'\1{self.ruble}'),)
        return expr

    def _positional_spaces(self, data, find, replace):
        """
        Helper method for `rep_positional_spaces` and `del_positional_spaces`
        expressions.
        """

        both = data.get('both', '')
        before = re.escape(data.get('before', '') + both)
        after = re.escape(data.get('after', '') + both)
        expr = []
        if before:
            expr.append((fr'{find}+(?=[{before}])', replace))
        if after:
            expr.append((fr'(?<=[{after}]){find}+', replace))
        return expr

    def _positional_spaces_before_currency_symbols(self):
        return [
            # add non-breaking space between value and currency symbol
            (
                fr'(?<=\d)({WHSP})*(?=[{"".join(CURRENCY_SYMBOLS)}]|{POLISH_ZLOTY})',
                NBSP,
            ),
            # replace non-breaking space to usual space after polish zloty currency,
            # that added in formulas transform (after ł)
            (fr'(?<=\d{NBSP}{POLISH_ZLOTY})({NBSP})', WHSP),
            # replace non-breaking space to usual space after currency symbols,
            # that uses in formulas transformations
            (
                fr'(?<=\d{NBSP}[{"".join(CURRENCY_SYMBOLS_USED_IN_FORMULAS)}])({NBSP})',
                WHSP,
            ),
        ]

    def expr_rep_positional_spaces(self):
        """
        Replaces whitespaces after and before certain symbols
        with non-breakable space.
        """

        expr = self._positional_spaces(self.rep_positional_spaces, WHSP, NBSP)
        expr += self._positional_spaces_before_currency_symbols()
        return expr

    def expr_del_positional_spaces(self):
        """
        Removes spaces before and after certain symbols.
        """

        expr = self._positional_spaces(self.del_positional_spaces, ANYSP, '')
        return expr

    def expr_rep_hyphen_between_short_words(self):
        """
        Replaces hyphen between two words. One of words should be at most 4 character.
        """
        hyphens = f'({HYPHEN}|{HYPHEN_MINUS})'
        return [
            (fr'(\b{self.words}{{1,4}}){hyphens}({self.words}+\b)', fr'\1{NBHYPHEN}\3'),
            (fr'(\b{self.words}+){hyphens}({self.words}{{1,4}}\b)', fr'\1{NBHYPHEN}\3'),
        ]


class RuExpressions(EnRuExpressions):
    """
    Reorder a `rep_hyphen_between_short_words` method
    to ensure `rep_positional_spaces_after_long_prepositions` works correctly because
    some long prepositions have hyphens.
    """

    expressions = (
        *without(EnRuExpressions.expressions, 'rep_hyphen_between_short_words'),
        'rep_positional_spaces_after_long_prepositions',
        'rep_hyphen_between_short_words',  # From EnRuExpressions
        'rep_positional_spaces_before_particles',
        'rep_positional_spaces_for_verbose_prepositions',
        'rep_custom_replacements',
    )

    def expr_rep_positional_spaces_after_long_prepositions(self):
        """
        Replaces whitespaces after and before certain words with non-breakable space.
        """
        expressions = []
        for word in RU_LONG_PREPOSITIONS:
            expressions.append((fr'(^{word}|\s{word}){WHSP}+', fr'\1{NBSP}'))
        return expressions

    def expr_rep_positional_spaces_before_particles(self):
        """Replaces whitespaces before particles with non-breakable space."""
        before_particles = [(fr'\s({word}\W)', fr'{NBSP}\1') for word in RU_PARTICLES]
        # Remove effect of `expr_pairs` for short particles
        short_particles = [word for word in RU_PARTICLES if len(word) < 3]
        after_short_particles = [
            (fr'(\W{word}){NBSP}+', fr'\1{WHSP}') for word in short_particles
        ]

        return [*before_particles, *after_short_particles]

    def expr_rep_positional_spaces_for_verbose_prepositions(self):
        """Replaces whitespaces in and after verbose prepositions
        with non-breakable spaces."""
        flags = re.M | re.S  # Skip re.I
        expressions = []
        for preposition_words in map(str.split, RU_VERBOSE_PREPOSITIONS):
            re_any_spaces_between_words = fr'{ANYSP}'.join(preposition_words)
            non_breaking_spaces_between_words = NBSP.join(preposition_words)

            exp_at_the_middle_of_sentence = (
                fr'(\s){re_any_spaces_between_words}\s',
                fr'\1{non_breaking_spaces_between_words}{NBSP}',
                flags,
            )
            expressions.append(exp_at_the_middle_of_sentence)
            exp_at_the_beginning_of_sentence = (
                fr'(^|\s){re_any_spaces_between_words.capitalize()}\s',
                fr'\1{non_breaking_spaces_between_words.capitalize()}{NBSP}',
                flags,
            )
            expressions.append(exp_at_the_beginning_of_sentence)

        return expressions

    def expr_rep_custom_replacements(self):
        """Replaces custom cases."""
        expressions = [
            (fr'\b{before}\b', fr'{after}') for before, after in RU_CUSTOM_REPLACEMENTS
        ]
        return expressions



from .chars import NBHYPHEN

RU_LONG_PREPOSITIONS = (
    'без',
    'перед',
    'при',
    'через',
    'над',
    'под',
    'про',
    'для',
    'около',
    'среди',
    'из-под',
    'из-за',
    'по-над',
)
RU_VERBOSE_PREPOSITIONS = ('несмотря на', 'в отличие от', 'в связи с')
RU_PARTICLES = ('же', 'ли', 'бы', 'б', 'уж')
RU_CUSTOM_REPLACEMENTS = (('когда-нибудь', f'когда{NBHYPHEN}нибудь'),)

CURRENCY_SYMBOLS = [
    '₽',
    '$',
    '€',
    '£',
    '₴',
    '¥',
    '₪',
    '₨',
    '₩',
    '₦',
    '฿',
    '₫',
    '₭',
    '៛',
    '₮',
    '₱',
    '₡',
    '₲',
    '₸',
    '₺',
    '₼',
    '₾',
    '¢',
    '₿',
    '﷼',
    'Ξ',
    'Ł',
]

CURRENCY_SYMBOLS_USED_IN_FORMULAS = ['Ξ', 'Ł']

# move polish zloty in standalone constant, because it uses two symbols for
# currency describing
POLISH_ZLOTY = 'Zł'



import re
from functools import wraps

__all__ = ('re_compile', 'idict', 'map_choices', 'splinter', 'without')


def re_compile(pattern, flags=re.I | re.U | re.M | re.S):  # noqa: E741
    """
    A shortcut to compile regex with predefined flags:
    :const:`re.I`, :const:`re.U`, :const:`re.M`, :const:`re.S`.

    :param str pattern: A string to compile pattern from.
    :param int flags: Python :mod:`re` module flags.

    >>> foo = re_compile('[a-z]')  # matches with 'test' and 'TEST'
    >>> bool(foo.match('TEST'))
    True
    >>> bar = re_compile('[a-z]', flags=0)  # doesn't match with 'TEST'
    >>> bool(bar.match('TEST'))
    False
    """

    return re.compile(pattern, flags)


class idict(dict):
    """
    Case-insensitive dictionary.

    :param mapping/iterable obj: An object to initialize new dictionary from
    :param `**kwargs`: ``key=value`` pairs to put in the new dictionary

    >>> foo = idict({'A': 0, 'b': 1}, bar=2)
    >>> foo['a'], foo['B'], foo['bAr']
    (0, 1, 2)

    .. caution::
        :class:`idict` is not a full-featured case-insensitive dictionary.
        As it's made for :func:`map_choices` and has limited functionality.
    """

    def __init__(self, obj=None, **kwargs):
        obj = dict(obj, **kwargs) if obj else kwargs
        normalized = {key.lower(): value for key, value in obj.items()}
        super().__init__(**normalized)

    def __getitem__(self, key):
        return super().__getitem__(key.lower())


def map_choices(data, group=r'({0})', dict_class=idict):
    """
    :class:`typus.processors.Expressions` helper.
    Builds regex pattern from the dictionary keys and maps them to values via
    replace function.

    :param mapping/iterable data: A pairs of (find, replace with) strings
    :param str group: A string to format in choices.
    :param class dict_class: A dictionary class to convert source data.
        By default :class:`idict` is used which is case-insensitive.
        In instance, to map  ``(c)`` and ``(C)`` to different values pass
        regular python :class:`dict`. Or if the order matters use
        :class:`collections.OrderedDict`

    :returns: A regex non-compiled pattern and replace function
    :rtype: tuple

    >>> import re
    >>> pattern, replace = map_choices({'a': 0, 'b': 1})
    >>> re.sub(pattern, replace, 'abc')
    '01c'
    """

    options = dict_class(data)
    choices = '|'.join(re.escape(x) for x in options)
    pattern = group.format(choices)

    def replace(match):
        key = match.group()
        return str(options[key])

    return pattern, replace


def splinter(delimiter):
    r"""
    :class:`typus.processors.EscapePhrases` helper.
    Almost like ``str.split()`` but handles delimiter escaping and strips
    spaces.

    :param str delimiter: String delimiter
    :raises ValueError: If delimiter is a slash or an empty space

    :returns: A list of stripped phrases splitted by the delimiter
    :rtype: list

    >>> split = splinter(',  ')  # strips this spaces
    >>> split('a, b,c ,  d\,e')  # and this ones too
    ['a', 'b', 'c', 'd,e']
    """

    delim = delimiter.strip(' \\')
    if not delim:
        raise ValueError('Delimiter can not be a slash or an empty space.')

    # Doesn't split escaped delimiters
    pattern = re.compile(fr'(?<!\\){re.escape(delim)}\s*')

    @wraps(splinter)
    def inner(phrases):
        # Deletes delimiter escaping and strips spaces
        return [x.replace('\\' + delim, delim).strip() for x in pattern.split(phrases)]

    return inner


def without(seq, *items):
    for value in seq:
        if value not in items:
            yield value



from functools import update_wrapper

from .chars import NBSP, NNBSP
from .utils import re_compile

__all__ = ('TypusCore',)


class TypusCore:
    """
    This class makes :mod:`typus.processors` and :mod:`typus.mixins` work
    together.
    """

    processors = ()
    expressions = ()
    re_nbsp = re_compile(f'[{NBSP}{NNBSP}]')

    def __init__(self):
        assert self.processors

        # Makes possible to decorate Typus.
        # updated=() skips __dict__ attribute
        update_wrapper(self, self.__class__, updated=())

        # Chains all processors into one single function
        self.process = sum(p(self) for p in reversed(self.processors))

    def __call__(self, text, debug=False, *args, **kwargs):
        text = text.strip()
        if not text:
            return ''

        # All the magic
        text = self.process(text, *args, **kwargs)

        # Makes nbsp visible
        if debug:
            return self.re_nbsp.sub('_', text)
        return text



from random import randrange

import pytest

from tj_text.typograph import en_typus, ru_typus
from tj_text.typograph.chars import HYPHEN, HYPHEN_MINUS, NBHYPHEN, NBSP
from tj_text.typograph.constants import (
    CURRENCY_SYMBOLS,
    CURRENCY_SYMBOLS_USED_IN_FORMULAS,
    POLISH_ZLOTY,
)
from tj_text.typograph.core import TypusCore
from tj_text.typograph.mixins import RuExpressions
from tj_text.typograph.processors import EscapeHtml, EscapePhrases, Expressions


@pytest.fixture()
def typus_factory():
    def factory(expression_mixin_cls, *, include=None, exclude=None):
        class Typus(expression_mixin_cls, TypusCore):
            processors = (EscapePhrases, EscapeHtml, Expressions)
            expressions = filter(
                lambda expr: (
                    (expr in include if include else True)
                    and (expr not in exclude if exclude else True)
                ),
                expression_mixin_cls.expressions,
            )

        return Typus()

    return factory


class TestEnRuExpressions:
    """Tests for typograph.mixins.EnRuExpressions class."""

    @pytest.mark.parametrize(
        'word_parts',
        [
            ('из', 'под'),
            ('когда', 'либо'),
            ('сим', 'карта'),
            ('QR', 'код'),
            ('где', 'нибудь'),
        ],
    )
    @pytest.mark.parametrize('hyphen', [HYPHEN, HYPHEN_MINUS])
    def test_expr_rep_hyphen_between_short_words(self, word_parts, hyphen):
        word = hyphen.join(word_parts)
        expected_word = NBHYPHEN.join(word_parts)
        text = (
            f'{word} проверим{word}дефис неразрывный {word} подставляется '
            f'корректно{word}между словами {word}.'
        )
        expected_text = (
            f'{expected_word} проверим{word}дефис неразрывный {expected_word} '
            f'подставляется корректно{word}между словами {expected_word}.'
        )

        result = en_typus(text)

        assert result == expected_text


class TestRuExpressions:
    """Tests for typograph.mixins.RuExpressions class."""

    def test_expressions_from_parent_class_are_in_child(self):
        for expression in ru_typus.expressions:
            assert expression in RuExpressions.expressions

    @pytest.mark.parametrize(
        'word',
        [
            'без',
            'перед',
            'при',
            'через',
            'над',
            'под',
            'про',
            'для',
            'около',
            'среди',
            'из-под',
            'из-за',
            'по-над',
        ],
    )
    def test_expr_rep_positional_spaces_after_long_prepositions(
        self, word, typus_factory
    ):
        typus = typus_factory(RuExpressions, exclude={'rep_hyphen_between_short_words'})
        text = (
            f'{word.capitalize()} проверяем{word} предлог {word} неразрывным '
            f'{word}пробелом {word}. {word.capitalize()} тоже.'
        )
        expected_result = (
            f'{word.capitalize()}{NBSP}проверяем{word} предлог {word}{NBSP}неразрывным '
            f'{word}пробелом {word}. {word.capitalize()}{NBSP}тоже.'
        )

        result = typus(text)

        assert result == expected_result

    @pytest.mark.parametrize('word', ['же', 'ли', 'бы', 'б', 'уж'])
    def test_expr_rep_positional_spaces_before_particles(self, word):
        text = f'Проверяем{word} частицу {word} неразрывным {word}пробелом {word}.'
        expected_result = (
            f'Проверяем{word} частицу{NBSP}{word} '
            f'неразрывным {word}пробелом{NBSP}{word}.'
        )

        result = ru_typus(text)

        assert result == expected_result

    @pytest.mark.parametrize(
        'original,expected',
        [
            ('несмотря на', f'несмотря{NBSP}на'),
            ('в отличие от', f'в{NBSP}отличие{NBSP}от'),
            ('в связи с', f'в{NBSP}связи{NBSP}с'),
        ],
    )
    def test_rep_positional_spaces_for_verbose_prepositions(self, original, expected):
        original_text = (
            f'{original.capitalize()} проверяем {original} неразрывным пробелом. '
            f'{original.capitalize()} даже в начале второго предложения.'
        )
        expected_result = (
            f'{expected.capitalize()}{NBSP}проверяем {expected}{NBSP}неразрывным '
            f'пробелом. '
            f'{expected.capitalize()}{NBSP}даже в{NBSP}начале второго предложения.'
        )

        result = ru_typus(original_text)

        assert result == expected_result

    @pytest.mark.parametrize(
        'original,expected', [('когда-нибудь', f'когда{NBHYPHEN}нибудь')]
    )
    def test_rep_custom_replacements(self, original, expected):
        original_text = (
            f'{original} проверим {original} слово-исключение '
            f'{original} будет сразу{original} заменено{original}.'
        )
        expected_result = (
            f'{expected} проверим {expected} слово-исключение '
            f'{expected} будет сразу{original} заменено{original}.'
        )

        result = ru_typus(original_text)

        assert result == expected_result

    @pytest.mark.parametrize(
        'currency',
        CURRENCY_SYMBOLS + CURRENCY_SYMBOLS_USED_IN_FORMULAS + [POLISH_ZLOTY],
    )
    @pytest.mark.parametrize('spaces_count', [0, 1, 10])
    def test_add_nbsp_between_number_and_currency_symbol_only(
        self, currency, spaces_count
    ):
        target_number = randrange(10000)
        original_text = f'{target_number}{" " * spaces_count}{currency} денег'
        expected_text = f'{target_number}{NBSP}{currency} денег'

        result = ru_typus(original_text)

        assert result == expected_text



NBSP = '\u00A0'
NNBSP = '\u202F'
WHSP = ' '
ANYSP = fr'[{WHSP}{NBSP}{NNBSP}]'

NDASH = '–'  # Ctrl + "minus" on keyboard \u2013
MDASH = '—'
MDASH_PAIR = NBSP + MDASH + WHSP

MINUS = '−'  # mathematical operator \u2212
HYPHEN_MINUS = '-'  # "minus" on keyboard \u002D
HYPHEN = '‐'  # \u2010
NBHYPHEN = '‑'  # \u2011

TIMES = '×'

LSQUO = '‘'  # left curly quote mark
RSQUO = '’'  # right curly quote mark/apostrophe
LDQUO = '“'  # left curly quote marks
RDQUO = '”'  # right curly quote marks
DLQUO = '„'  # double low curly quote mark
LAQUO = '«'  # left angle quote marks
RAQUO = '»'  # right angle quote marks

SPRIME = '′'
DPRIME = '″'



from .core import TypusCore
from .mixins import EnQuotes, EnRuExpressions, RuExpressions, RuQuotes
from .processors import EscapeHtml, EscapePhrases, Expressions

__all__ = ('en_typus', 'ru_typus')


class EnTypus(EnQuotes, EnRuExpressions, TypusCore):
    processors = (EscapePhrases, EscapeHtml, Expressions)


class RuTypus(RuQuotes, RuExpressions, TypusCore):
    processors = (EscapePhrases, EscapeHtml, Expressions)


en_typus, ru_typus = EnTypus(), RuTypus()



from functools import update_wrapper, wraps
from itertools import count, cycle

from .chars import DLQUO, LAQUO, LDQUO, LSQUO, RAQUO, RDQUO, RSQUO
from .utils import re_compile

__all__ = ('EscapePhrases', 'EscapeHtml', 'Quotes', 'Expressions')


def tail_processor(text, *args, **kwargs):
    return text


class BaseProcessor:
    """
    Processors are the core of Typus. See subclasses for examples.
    """

    def __init__(self, typus):
        # Makes possible to decorate processor
        update_wrapper(self, self.__class__, updated=())

        # Stores Typus to access it's configuration
        self.typus = typus

    def __call__(self, typus):
        raise NotImplementedError

    def __radd__(self, other):
        return self(other or tail_processor)


class EscapePhrases(BaseProcessor):
    """
    Escapes phrases which should never be processed.

    >>> en_typus('Typus turns `(c)` into "(c)"', escape_phrases=['`(c)`'])
    'Typus turns `(c)` into “©”'

    Also there is a little helper :func:`typus.utils.splinter` which should
    help you to split string into the phrases.
    """

    placeholder = '{{#phrase{0}#}}'

    def __call__(self, func):
        @wraps(self, updated=())
        def inner(text, *args, **kwargs):
            storage = []
            counter = count()
            escaped = self._save_values(text, storage, counter, **kwargs)

            # Runs typus
            processed = func(escaped, *args, **kwargs)
            if not storage:
                return processed

            restored = self._restore_values(processed, storage, **kwargs)
            return restored

        return inner

    def _save_values(self, text, storage, counter, escape_phrases=(), **kwargs):
        for phrase in escape_phrases:
            if not phrase.strip():
                continue
            key = self.placeholder.format(next(counter))
            text = text.replace(phrase, key)
            storage.append((key, phrase))
        return text

    def _restore_values(self, text, storage, **kwargs):
        """
        Puts data into the text in reversed order.
        It's important to loop over and restore text step by step
        because some 'stored' chunks may contain keys to other ones.
        """
        for key, value in reversed(storage):
            text = text.replace(key, value)
        return text


class EscapeHtml(EscapePhrases):
    """
    Extracts html tags and puts them back after.

    >>> en_typus('Typus turns <code>(c)</code> into "(c)"')
    'Typus turns <code>(c)</code> into “©”'

    .. caution::
        Doesn't support nested ``<code>`` tags.
    """

    placeholder = '{{#html{0}#}}'
    skiptags = 'head|iframe|pre|code|script|style|video|audio|canvas'
    patterns = (
        re_compile(fr'(<)({skiptags})(.*?>.*?</\2>)'),
        # Doctype, xml, closing tag, any tag
        re_compile(r'(<[\!\?/]?[a-z]+.*?>)'),
        # Comments
        re_compile(r'(<\!\-\-.*?\-\->)'),
    )

    def _save_values(self, text, storage, counter, **kwargs):
        for pattern in self.patterns:
            text = pattern.sub(self._replace(storage, counter), text)
        return text

    def _replace(self, storage, counter):
        def inner(match):
            key = self.placeholder.format(next(counter))
            html = ''.join(match.groups())
            storage.append((key, html))
            return key

        return inner


class Quotes(BaseProcessor):
    """
    Replaces regular quotes with typographic ones.
    Supports any level nesting, but doesn't work well with minutes ``1'``
    and inches ``1"`` within the quotes, that kind of cases are ignored.
    Use it with :class:`typus.mixins.RuQuotes` or
    :class:`typus.mixins.EnQuotes` or provide Typus attributes
    ``loq, roq, leq, req`` with custom quotes.

    >>> en_typus('Say "what" again!')
    'Say “what” again!'
    """

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        # Odd and even levels: left, right
        self.loq, self.roq = self.typus.loq, self.typus.roq
        self.leq, self.req = self.typus.leq, self.typus.req

        # Pairs of odd and even quotes. Already *switched* in one dimension.
        # See :meth:`_switch_nested` for more help.
        self.switch = (self.loq + self.req, self.leq + self.roq)

        # Replaces all quotes with `'`
        quotes = ''.join((LSQUO, RSQUO, LDQUO, RDQUO, DLQUO, LAQUO, RAQUO))
        self.re_normalize = re_compile(fr'[{quotes}]')

        # Matches nested quotes (with no quotes within)
        # and replaces with odd level quotes
        self.re_normal = re_compile(
            # No words before
            r'(?<!\w)'
            # Starts with quote
            r'(["\'])'
            r'(?!\s)'
            # Everything but quote inside
            r'((?!\1).+?)'
            r'(?!\s)'
            # Ends with same quote from the beginning
            r'\1'
            # No words afterwards
            r'(?!\w)'
        )
        self.re_normal_replace = fr'{self.loq}\2{self.roq}'

        # Matches with typo quotes
        self.re_nested = re_compile(fr'({self.loq}|{self.roq})')

    def __call__(self, func):
        @wraps(self, updated=())
        def inner(text, *args, **kwargs):
            # Normalizes editor's quotes to double one
            normalized = self.re_normalize.sub('\'', text)

            # Replaces normalized quotes with first level ones, starting
            # from inner pairs, moves to sides
            nested = 0
            while True:
                normalized, replaced = self.re_normal.subn(
                    self.re_normal_replace, normalized
                )
                if not replaced:
                    break
                nested += 1

            # Saves some cpu :)
            # Most cases are about just one level quoting
            if nested < 2:
                return func(normalized, *args, **kwargs)

            # At this point all quotes are of odd type, have to fix it
            switched = self._switch_nested(normalized)
            return func(switched, *args, **kwargs)

        return inner

    def _switch_nested(self, text):
        """
        Switches nested quotes to another type.
        This function stored in a separate method to make possible to mock it
        in tests to make sure it doesn't called without special need.
        """

        # Stores a cycled pairs of possible quotes. Every other loop it's
        # switched to provide *next* type of a given quote
        quotes = cycle(self.switch)

        def replace(match):
            # Since only odd quotes are matched, comparison is the way to
            # choose whether it's left or right one of type should be returned.
            # As the first quote is the left one, makes negative equal which
            # return false, i.e. zero index
            return next(quotes)[match.group() != self.loq]

        return self.re_nested.sub(replace, text)


class Expressions(BaseProcessor):
    r"""
    Provides regular expressions support. Looks for ``expressions`` list
    attribute in Typus with expressions name, compiles and runs them on every
    Typus call.

    >>> from typus.core import TypusCore
    >>> from typus.processors import Expressions
    ...
    >>> class MyExpressionsMixin:
    ...     def expr_bold_price(self):
    ...         expr = (
    ...             (r'(\$\d+)', r'<b>\1</b>'),
    ...         )
    ...         return expr
    ...
    >>> class MyTypus(MyExpressionsMixin, TypusCore):
    ...     expressions = ('bold_price', )  # no prefix `expr_`!
    ...     processors = (Expressions, )
    ...
    >>> my_typus = MyTypus()  # `expr_bold_price` is compiled and stored
    >>> my_typus('Get now just for $1000!')
    'Get now just for <b>$1000</b>!'

    .. note::
        *Expression* is a pair of regex and replace strings. Regex strings are
        compiled with :func:`typus.utils.re_compile` with a bunch of flags:
        unicode, case-insensitive, etc. If that doesn't suit for you pass your
        own flags as a third member of the tuple: ``(regex, replace, re.I)``.
    """

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        # Compiles expressions
        self.compiled_exprs = [
            (re_compile(*group[::2]), group[1])
            for name in self.typus.expressions
            for group in getattr(self.typus, 'expr_' + name)()
        ]

    def __call__(self, func):
        @wraps(self, updated=())
        def inner(text, *args, **kwargs):
            # Applies expressions
            for expr, repl in self.compiled_exprs:
                text = expr.sub(repl, text)
            text = func(text, *args, **kwargs)
            return text

        return inner
